use std::{
    fs::File,
    io::{BufReader, BufWriter, Read, Write},
    path::Path,
};

#[derive(Debug, Clone, Copy)]
pub struct Color {
    pub red: u8,
    pub green: u8,
    pub blue: u8,
}

impl Color {
    pub fn grey(&self) -> Self {
        let r = self.red as f64;
        let g = self.green as f64;
        let b = self.blue as f64;
        let grey = (0.2125 * r + 0.7154 * g + 0.0721 * b) as u8;
        Self {
            red: grey,
            green: grey,
            blue: grey,
        }
    }

    pub fn new_from_yuv(
        y: u8,
        u: u8,
        v: u8,
    ) -> Self {
        let yd = 1.1640625 * (y as f64 - 16.0);
        let ud = u as f64 - 128.0;
        let vd = v as f64 - 128.0;
        let c0 = 1.59765625;
        let c1 = -0.390625;
        let c2 = -0.8125;
        let c3 = 2.015625;
        let red = (yd + c0 * vd).clamp(0.0, 255.0) as u8;
        let green = (yd + c1 * ud + c2 * vd).clamp(0.0, 255.0) as u8;
        let blue = (yd + c3 * ud).clamp(0.0, 255.0) as u8;
        Self { red, green, blue }
    }
}

pub struct Image {
    name: String,
    width: usize,
    height: usize,
    pixels: Vec<Color>,
}

impl Image {
    pub fn new_unicolor(
        name: String,
        width: usize,
        height: usize,
        color: Color,
    ) -> Self {
        Self {
            name,
            width,
            height,
            pixels: vec![color; width * height],
        }
    }

    pub fn new_random(
        name: String,
        width: usize,
        height: usize,
    ) -> Self {
        use rand::Rng;
        let mut rng = rand::thread_rng();
        let mut pixels = Vec::with_capacity(width * height);
        for _ in 0..width * height {
            pixels.push(Color {
                red: rng.gen(),
                green: rng.gen(),
                blue: rng.gen(),
            })
        }
        Self {
            name,
            width,
            height,
            pixels,
        }
    }

    pub fn name(&self) -> &str {
        &self.name
    }

    pub fn size(&self) -> (usize, usize) {
        (self.width, self.height)
    }

    pub fn save(
        &self,
        path: &str,
    ) -> Result<(), Box<dyn std::error::Error>> {
        let mut output = BufWriter::new(File::create(path)?);
        writeln!(output, "P3")?;
        writeln!(output, "{} {}", self.width, self.height)?;
        writeln!(output, "255")?;
        for c in self.pixels.iter() {
            writeln!(output, "{} {} {}", c.red, c.green, c.blue)?;
        }
        Ok(())
    }

    pub fn new_from_rgb(
        path: &str
    ) -> Result<Self, Box<dyn std::error::Error>> {
        let content = std::fs::read(path)?;
        let mut words = std::str::from_utf8(&content)?
            .lines()
            .map(|l| l.find('#').map_or(l, |pos| &l[0..pos]))
            .flat_map(|l| l.split_whitespace())
            .filter(|w| !w.is_empty());
        let p3 = words.next().ok_or("missing P3")?;
        if p3 != "P3" {
            Err(format!("unexpected {:?}", p3))?;
        }
        let width = words.next().ok_or("missing width")?.parse::<usize>()?;
        let height =
            words.next().ok_or("missing height")?.parse::<usize>()?;
        let c255 = words.next().ok_or("missing 255")?;
        if c255 != "255" {
            Err(format!("unexpected {:?}", c255))?;
        }
        let mut pixels = Vec::with_capacity(width * height);
        for _ in 0..width * height {
            let red = words.next().ok_or("missing red")?.parse::<u8>()?;
            let green = words.next().ok_or("missing green")?.parse::<u8>()?;
            let blue = words.next().ok_or("missing blue")?.parse::<u8>()?;
            pixels.push(Color { red, green, blue })
        }
        Ok(Self {
            name: Self::make_name(path),
            width,
            height,
            pixels,
        })
    }

    pub fn grey(&mut self) {
        self.name.push_str("_grey");
        for c in self.pixels.iter_mut() {
            *c = c.grey();
        }
    }

    pub fn new_from_yuv(
        path: &str
    ) -> Result<Self, Box<dyn std::error::Error>> {
        let mut input = BufReader::new(File::open(path)?);
        let width = 640;
        let height = 480;
        let mut pixels = Vec::with_capacity(width * height);
        for _ in 0..width * height / 2 {
            let mut yuyv = [0_u8; 4];
            input.read_exact(&mut yuyv)?;
            let [y0, u, y1, v] = yuyv;
            pixels.push(Color::new_from_yuv(y0, u, v));
            pixels.push(Color::new_from_yuv(y1, u, v));
        }
        Ok(Self {
            name: Self::make_name(path),
            width,
            height,
            pixels,
        })
    }

    fn make_name(path: &str) -> String {
        let p = Path::new(path).file_stem().unwrap();
        Path::new(p)
            .file_name()
            .unwrap()
            .to_str()
            .unwrap()
            .to_owned()
    }
}
